@@ -58,8 +58,6 @@
 #include "errorgui.h"
 
 
-static const gchar RCSid[]="$Id: slave.c,v 1.104 2001/10/28 19:32:04 jirka Exp $";
-
 /* Some per slave globals */
 static GdmDisplay *d;
 static gchar *login = NULL;
@@ -417,6 +415,10 @@ gdm_slave_run (GdmDisplay *display)
 	}
     }
 
+    if (SERVER_IS_LOCAL (d)) {
+	    gdm_slave_send (GDM_SOP_START_NEXT_LOCAL, FALSE);
+    }
+
     /* something may have gone wrong, try xfailed, if local (non-flexi),
      * the toplevel loop of death will handle us */ 
     if (d->dsp == NULL) {
@@ -869,7 +871,9 @@ run_pictures (void)
 	FILE *fp;
 	char *cfgdir;
 
+	response = NULL;
 	for (;;) {
+		g_free (response);
 		response = gdm_slave_greeter_ctl (GDM_NEEDPIC, "");
 		if (ve_string_empty (response)) {
 			g_free (response);
@@ -1041,6 +1045,7 @@ run_pictures (void)
 		seteuid (0);
 		setegid (GdmGroupId);
 	}
+	g_free (response);
 }
 
 static void
@@ -1251,30 +1256,28 @@ gdm_slave_greeter (void)
 }
 
 void
-gdm_slave_send_num (const char *opcode, long num)
+gdm_slave_send (const char *str, gboolean wait_for_usr2)
 {
 	char *msg;
 	int fd;
 	char *fifopath;
 
-	gdm_debug ("Sending %s == %ld for slave %ld",
-		   opcode,
-		   (long)num,
-		   (long)getpid ());
+	gdm_debug ("Sending %s", str);
 
-	gdm_got_usr2 = FALSE;
+	if (wait_for_usr2)
+		gdm_got_usr2 = FALSE;
 
 	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
-
 	fd = open (fifopath, O_WRONLY);
+	g_free (fifopath);
+
 	/* eek */
 	if (fd < 0) {
-		gdm_error (_("%s: Can't open fifo!"), "gdm_send_num");
+		gdm_error (_("%s: Can't open fifo!"), "gdm_slave_send");
 		return;
 	}
 
-	msg = g_strdup_printf ("\n%s %ld %ld\n", opcode,
-			       (long)getpid (), (long)num);
+	msg = g_strdup_printf ("\n%s\n", str);
 
 	write (fd, msg, strlen (msg));
 
@@ -1282,16 +1285,34 @@ gdm_slave_send_num (const char *opcode, long num)
 
 	close (fd);
 
-	if ( ! gdm_got_usr2)
+	if (wait_for_usr2 &&
+	    ! gdm_got_usr2)
 		sleep (10);
 }
 
+void
+gdm_slave_send_num (const char *opcode, long num)
+{
+	char *msg;
+
+	gdm_debug ("Sending %s == %ld for slave %ld",
+		   opcode,
+		   (long)num,
+		   (long)getpid ());
+
+
+	msg = g_strdup_printf ("%s %ld %ld", opcode,
+			       (long)getpid (), (long)num);
+
+	gdm_slave_send (msg, TRUE);
+
+	g_free (msg);
+}
+
 void
 gdm_slave_send_string (const char *opcode, const char *str)
 {
 	char *msg;
-	int fd;
-	char *fifopath;
 
 	/* Evil!, all this for debugging? */
 	if (GdmDebug) {
@@ -1306,28 +1327,12 @@ gdm_slave_send_string (const char *opcode, const char *str)
 				   (long)getpid ());
 	}
 
-	gdm_got_usr2 = FALSE;
-
-	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
-
-	fd = open (fifopath, O_WRONLY);
-	/* eek */
-	if (fd < 0) {
-		gdm_error (_("%s: Can't open fifo!"), "gdm_slave_send_string");
-		return;
-	}
-
-	msg = g_strdup_printf ("\n%s %ld %s\n", opcode,
+	msg = g_strdup_printf ("%s %ld %s", opcode,
 			       (long)getpid (), ve_sure_string (str));
 
-	write (fd, msg, strlen (msg));
+	gdm_slave_send (msg, TRUE);
 
 	g_free (msg);
-
-	close (fd);
-
-	if ( ! gdm_got_usr2)
-		sleep (10);
 }
 
 static void
@@ -1352,6 +1357,8 @@ send_chosen_host (GdmDisplay *disp, const char *hostname)
 
 	fd = open (fifopath, O_WRONLY);
 
+	g_free (fifopath);
+
 	/* eek */
 	if (fd < 0) {
 		gdm_error (_("%s: Can't open fifo!"), "send_chosen_host");
@@ -2483,13 +2490,22 @@ gdm_slave_child_handler (int sig)
 		gdm_server_stop (d);
 		gdm_verify_cleanup (d);
 
-		if (WIFEXITED (status)) {
+		/* The greeter is only allowed to pass back these
+		 * exit codes, else we'll just remanage */
+		if (WIFEXITED (status) &&
+		    (WEXITSTATUS (status) == DISPLAY_ABORT ||
+		     WEXITSTATUS (status) == DISPLAY_REBOOT ||
+		     WEXITSTATUS (status) == DISPLAY_HALT ||
+		     WEXITSTATUS (status) == DISPLAY_SUSPEND ||
+		     WEXITSTATUS (status) == DISPLAY_RESTARTGDM)) {
 			_exit (WEXITSTATUS (status));
 		} else {
 			_exit (DISPLAY_REMANAGE);
 		}
 	} else if (pid != 0 && pid == d->sesspid) {
 		d->sesspid = 0;
+	} else if (pid != 0 && pid == d->chooserpid) {
+		d->chooserpid = 0;
 	} else if (pid != 0 && pid == d->servpid) {
 		d->servstat = SERVER_DEAD;
 		d->servpid = 0;
@@ -2679,7 +2695,8 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
 		       struct passwd *pwent)
 {
     pid_t pid;
-    gchar *script, *defscript, *scr;
+    gchar *script, *defscript;
+    const char *scr;
     gchar **argv;
     gint status;
     int i;
@@ -2690,12 +2707,15 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
     script = g_strconcat (dir, "/", d->name, NULL);
     defscript = g_strconcat (dir, "/Default", NULL);
 
-    if (! access (script, R_OK|X_OK))
-	scr = script;
-    else if (! access (defscript, R_OK|X_OK)) 
-	scr = defscript;
-    else
-	return EXIT_SUCCESS;
+    if (! access (script, R_OK|X_OK)) {
+	    scr = script;
+    } else if (! access (defscript, R_OK|X_OK))  {
+	    scr = defscript;
+    } else {
+	    g_free (script);
+	    g_free (defscript);
+	    return EXIT_SUCCESS;
+    }
 
     pid = gdm_fork_extra ();
 
@@ -2743,12 +2763,17 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
 	_exit (EXIT_SUCCESS);
 	    
     case -1:
+	g_free (script);
+	g_free (defscript);
 	syslog (LOG_ERR, _("gdm_slave_exec_script: Can't fork script process!"));
 	return EXIT_SUCCESS;
 	
     default:
 	gdm_wait_for_extra (&status);
 
+	g_free (script);
+	g_free (defscript);
+
 	if (WIFEXITED (status))
 	    return WEXITSTATUS (status);
 	else
